from collections import OrderedDict
from inspect import Parameter, signature

import pytest

from attr import define

from incant import Incanter


def test_simple_dep(incanter: Incanter):
    def func(dep1) -> int:
        return dep1 + 1

    with pytest.raises(TypeError):
        incanter.invoke(func)
    assert signature(incanter.prepare(func)).parameters == {
        "dep1": Parameter("dep1", Parameter.POSITIONAL_OR_KEYWORD)
    }

    incanter.register_hook(lambda p: p.name == "dep1", lambda: 2)
    assert incanter.invoke(func) == 3

    assert signature(incanter.prepare(func)).parameters == {}
    assert signature(incanter.prepare(func)).return_annotation is int


def test_nested_deps(incanter: Incanter):
    incanter.register_hook(lambda p: p.name == "dep1", lambda dep2: dep2 + 1)
    incanter.register_hook(lambda p: p.name == "dep2", lambda: 2)

    def func(dep1):
        return dep1 + 1

    assert signature(incanter.prepare(func)).parameters == {}
    assert incanter.invoke(func) == 4


def test_nested_partial_deps(incanter: Incanter):
    incanter.register_hook(
        lambda p: p.name == "dep1", lambda dep2, input: dep2 + input + 1
    )
    incanter.register_hook(lambda p: p.name == "dep2", lambda: 2)

    def func(dep1):
        return dep1 + 1

    assert signature(incanter.prepare(func)).parameters == {
        "input": Parameter("input", Parameter.POSITIONAL_OR_KEYWORD)
    }
    assert incanter.invoke(func, 1) == 5


def test_nested_partial_deps_with_args(incanter: Incanter):
    incanter.register_hook(
        lambda p: p.name == "dep1", lambda dep2, input: dep2 + input + 1
    )
    incanter.register_hook(lambda p: p.name == "dep2", lambda: 2)

    def func(dep1, input2: float) -> float:
        return dep1 + 1 + input2

    assert signature(incanter.prepare(func)).parameters == OrderedDict(
        [
            ("input", Parameter("input", Parameter.POSITIONAL_OR_KEYWORD)),
            (
                "input2",
                Parameter("input2", Parameter.POSITIONAL_OR_KEYWORD, annotation=float),
            ),
        ]
    )
    assert incanter.invoke(func, 1, 5.0) == 10.0


def test_nested_partial_deps_with_coalesce(incanter: Incanter):
    @incanter.register_by_type
    def dep(arg: float) -> str:
        return str(arg + 1)

    def fn(arg: str):
        return "1.0" + arg

    assert incanter.invoke(fn, 1.0) == "1.02.0"


def test_shared_deps(incanter: Incanter):
    incanter.register_by_name(lambda dep2, input: dep2 + input + 1, name="dep1")
    incanter.register_by_name(lambda: 2, name="dep2")

    def func(dep1, input: float) -> float:
        return dep1 + 1 + input

    assert signature(incanter.prepare(func)).parameters == OrderedDict(
        [
            (
                "input",
                Parameter("input", Parameter.POSITIONAL_OR_KEYWORD, annotation=float),
            ),
        ]
    )
    assert incanter.invoke(func, 5.0) == 14.0


def test_shared_deps_type_from_dep(incanter: Incanter):
    """The parameter type definition comes from the dependency."""

    @incanter.register_by_name
    def dep1(dep2, input: float):
        return dep2 + input + 1

    incanter.register_by_name(lambda: 2, name="dep2")

    def func(dep1, input) -> float:
        return dep1 + 1 + input

    assert signature(incanter.prepare(func)).parameters == OrderedDict(
        [
            (
                "input",
                Parameter("input", Parameter.POSITIONAL_OR_KEYWORD, annotation=float),
            ),
        ]
    )
    assert incanter.invoke(func, 5.0) == 14.0


def test_shared_deps_incompatible(incanter: Incanter):
    """Type incompatibilities are raised."""

    def dep1(dep2, input: str) -> str:
        return dep2 + input + "1"

    incanter.register_hook(lambda p: p.name == "dep1", dep1)
    incanter.register_hook(lambda p: p.name == "dep2", lambda: 2)

    def func(dep1, input: float) -> float:
        return dep1 + 1 + input

    with pytest.raises(Exception):
        incanter.prepare(func)

    with pytest.raises(Exception):
        incanter.invoke(func, 5.0)


def test_class_deps(incanter: Incanter):
    @define
    class Dep:
        a: int

    incanter.register_hook(lambda p: p.name == "dep", Dep)
    assert incanter.invoke(lambda dep: dep.a + 1, a=1)

    assert signature(incanter.prepare(lambda dep: dep.a + 1)).parameters == OrderedDict(
        [
            (
                "a",
                Parameter("a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
            ),
        ]
    )


def test_no_return_type(incanter: Incanter):
    """Registering by type with no return type is an error."""

    def dep():
        return 1

    with pytest.raises(Exception):
        incanter.register_by_type(dep)


def test_optional_arg(incanter: Incanter):
    """Registering by type with no return type is an error."""

    @incanter.register_by_name
    def dep(i=1):
        return i

    assert incanter.invoke(lambda dep: dep + 1) == 2
    assert incanter.invoke(lambda dep: dep + 1, 2) == 3
    assert signature(incanter.prepare(lambda dep: dep + 1)).parameters == {
        "i": Parameter("i", Parameter.POSITIONAL_OR_KEYWORD, default=1)
    }


def test_same_type_arg_coalescing(incanter: Incanter):
    @incanter.register_by_name
    def dep(i: int) -> int:
        return 1

    def fn(i: int, dep: int) -> int:
        return i + dep

    assert incanter.invoke(fn, 1) == 2



from functools import lru_cache
from inspect import Parameter, Signature, iscoroutinefunction, signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from attr import Factory, define, field, frozen

from ._codegen import (
    LocalVarFactory,
    ParameterDep,
    compile_incant_wrapper,
    compile_invoke,
)


_type = type


R = TypeVar("R")


@frozen
class FactoryDep:
    factory: Callable  # The fn to call.
    arg_name: str  # The name of the param this is fulfulling.


Dep = Union[FactoryDep, ParameterDep]


PredicateFn = Callable[[Parameter], bool]


def is_subclass(type, superclass) -> bool:
    try:
        return issubclass(type, superclass)
    except Exception:
        return False


@frozen
class Hook:
    predicate: PredicateFn
    factory: Optional[Callable[[Parameter], Callable]]

    @classmethod
    def for_name(cls, name: str, hook: Optional[Callable]):
        return cls(lambda p: p.name == name, None if hook is None else (lambda _: hook))  # type: ignore

    @classmethod
    def for_type(cls, type: Any, hook: Optional[Callable]):
        """Register by exact type (subclasses won't match)."""
        return cls(
            lambda p: p.annotation == type,
            None if hook is None else lambda _: hook,  # type: ignore
        )


@define
class Incanter:
    hook_factory_registry: List[Hook] = Factory(list)
    _invoke_cache: Callable = field(
        init=False,
        default=Factory(
            lambda self: lru_cache(None)(self._gen_invoke), takes_self=True
        ),
    )
    _incant_cache: Callable = field(
        init=False,
        default=Factory(
            lambda self: lru_cache(None)(self._gen_incant), takes_self=True
        ),
    )

    def prepare(
        self,
        fn: Callable[..., R],
        hooks: Sequence[Hook] = (),
        is_async: Optional[bool] = None,
    ) -> Callable[..., R]:
        return self._invoke_cache(fn, tuple(hooks), is_async)

    def invoke(self, fn: Callable[..., R], *args, **kwargs) -> R:
        return self.prepare(fn, is_async=False)(*args, **kwargs)

    async def ainvoke(self, fn: Callable[..., Awaitable[R]], *args, **kwargs) -> R:
        return await self.prepare(fn, is_async=True)(*args, **kwargs)

    def incant(self, fn: Callable[..., R], *args, **kwargs) -> R:
        """Invoke `fn` the best way we can."""
        return self._incant(fn, args, kwargs)

    async def aincant(self, fn: Callable[..., Awaitable[R]], *args, **kwargs) -> R:
        """Invoke async `fn` the best way we can."""
        return await self._incant(fn, args, kwargs, is_async=True)

    def register_by_name(
        self, fn: Optional[Callable] = None, *, name: Optional[str] = None
    ):
        """
        Register a factory to be injected by name. Can also be used as a decorator.

        If the name is not provided, the name of the factory will be used.
        """
        if fn is None:
            # Decorator
            return lambda fn: self.register_by_name(fn, name=name)

        if name is None:
            name = fn.__name__
        self.register_hook(lambda p: p.name == name, fn)

    def register_by_type(self, fn: Union[Callable, Type], type: Optional[Type] = None):
        """
        Register a factory to be injected by type. Can also be used as a decorator.

        If the type is not provided, the return annotation from the
        factory will be used.
        """
        if type is None:
            if isinstance(fn, _type):
                type_to_reg = fn
            else:
                sig = signature(fn)
                type_to_reg = sig.return_annotation
                if type_to_reg is Signature.empty:
                    raise Exception("No return type found, provide a type.")
        else:
            type_to_reg = type
        self.register_hook(lambda p: is_subclass(p.annotation, type_to_reg), fn)

    def register_hook(self, predicate: PredicateFn, factory: Callable):
        self.register_hook_factory(predicate, lambda _: factory)

    def register_hook_factory(
        self, predicate: PredicateFn, hook_factory: Callable[[Parameter], Callable]
    ):
        self.hook_factory_registry.insert(0, Hook(predicate, hook_factory))
        self._invoke_cache.cache_clear()  # type: ignore
        self._incant_cache.cache_clear()  # type: ignore

    def _incant(
        self,
        fn: Callable,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        is_async: bool = False,
    ):
        """The shared entrypoint for ``incant`` and ``aincant``."""

        pos_args_types = tuple([a.__class__ for a in args])
        kwargs_by_name_and_type = frozenset(
            [(k, v.__class__) for k, v in kwargs.items()]
        )
        wrapper = self._incant_cache(
            fn, pos_args_types, kwargs_by_name_and_type, is_async
        )

        return wrapper(*args, **kwargs)

    def _gen_incant_plan(
        self, fn, pos_args_types: Tuple, kwargs: Set[Tuple[str, Any]], is_async=False
    ) -> List[Union[int, str]]:
        """Generate a plan to invoke `fn`, potentially using `args` and `kwargs`."""
        pos_arg_plan: List[Union[int, str]] = []
        pos_args_by_type = {a: ix for ix, a in enumerate(pos_args_types)}
        kwarg_names = {kw[0] for kw in kwargs}
        sig = signature(fn)
        for arg_name, arg in sig.parameters.items():
            if (
                arg.annotation is not Signature.empty
                and (arg_name, arg.annotation) in kwargs
            ):
                pos_arg_plan.append(arg_name)
            elif (
                arg.annotation is not Signature.empty
                and arg.annotation in pos_args_by_type
            ):
                pos_arg_plan.append(pos_args_by_type[arg.annotation])
            elif arg_name in kwarg_names:
                pos_arg_plan.append(arg_name)
            else:
                raise TypeError(f"Cannot fulfil argument {arg_name}")
        return pos_arg_plan

    def _gen_incant(
        self,
        fn: Callable,
        pos_args_types: Tuple,
        kwargs_by_name_and_type: Set,
        is_async: Optional[bool] = False,
    ) -> Callable:
        plan = self._gen_incant_plan(
            fn, pos_args_types, kwargs_by_name_and_type, is_async=is_async
        )
        incant = compile_incant_wrapper(
            fn, plan, len(pos_args_types), len(kwargs_by_name_and_type)
        )
        return incant

    def _gen_dep_tree(
        self, fn: Callable, additional_hooks: Sequence[Hook]
    ) -> List[Tuple[Callable, List[Dep]]]:
        """Generate the dependency tree for `fn`."""
        to_process = [fn]
        final_nodes: List[Tuple[Callable, List[Dep]]] = []
        hooks = list(additional_hooks) + self.hook_factory_registry
        while to_process:
            _nodes = to_process
            to_process = []
            for node in _nodes:
                sig = signature(node)
                dependents: List[Union[ParameterDep, FactoryDep]] = []
                for name, param in sig.parameters.items():
                    if (
                        node is not fn
                        and param.default is not Signature.empty
                        and param.kind is Parameter.KEYWORD_ONLY
                    ):
                        # Do not expose optional kw-only params of dependencies.
                        continue
                    param_type = param.annotation
                    for hook in hooks:
                        if hook.predicate(param):
                            # Match!
                            if hook.factory is None:
                                dependents.append(
                                    ParameterDep(name, param_type, param.default)
                                )
                            else:
                                factory = hook.factory(param)
                                to_process.append(factory)
                                dependents.append(FactoryDep(factory, name))
                            break
                    else:
                        dependents.append(ParameterDep(name, param_type, param.default))
                final_nodes.insert(0, (node, dependents))
        return final_nodes

    def _gen_invoke(
        self,
        fn: Callable,
        hooks: Tuple[Hook, ...] = (),
        is_async: Optional[bool] = False,
    ):
        dep_tree = self._gen_dep_tree(fn, hooks)

        local_vars: List[LocalVarFactory] = []

        # is_async = None means autodetect
        if is_async is None:
            is_async = any(iscoroutinefunction(factory) for factory, _ in dep_tree)
        # All non-parameter deps become local vars.
        for factory, deps in dep_tree[:-1]:
            if not is_async and iscoroutinefunction(factory):
                raise Exception(
                    f"The function would be a coroutine because of {factory}, use `ainvoke` instead"
                )
            local_vars.append(
                LocalVarFactory(
                    factory,
                    [
                        dep.factory if isinstance(dep, FactoryDep) else dep
                        for dep in deps
                    ],
                )
            )

        outer_args = [
            dep for node in dep_tree for dep in node[1] if isinstance(dep, ParameterDep)
        ]
        # We need to do a pass over the outer args to consolidate duplicates.
        per_outer_arg: dict[str, List[ParameterDep]] = {}
        for arg in outer_args:
            per_outer_arg.setdefault(arg.arg_name, []).append(arg)

        outer_args.clear()
        for arg_name, args in per_outer_arg.items():
            if len(args) == 1:
                arg_type = args[0].type
                arg_default = args[0].default
            else:
                # If there are multiple competing argument defs,
                # we need to pick a winning type.
                arg_type = Signature.empty
                arg_default = Signature.empty
                for arg in args:
                    try:
                        arg_type = _reconcile_types(arg_type, arg.type)
                    except Exception:
                        raise Exception(
                            f"Unable to reconcile types {arg_type} and {arg.type} for argument {arg_name}"
                        )
                    if arg.default is not Signature.empty:
                        arg_default = arg.default
            outer_args.append(ParameterDep(arg_name, arg_type, arg_default))

        # outer_args need to be sorted by the presence of a default value
        outer_args.sort(key=lambda a: a.default is not Signature.empty)

        fn_factory_args = []
        for dep in dep_tree[-1][1]:
            if isinstance(dep, FactoryDep):
                fn_factory_args.append(dep.arg_name)

        return compile_invoke(
            fn, fn_factory_args, outer_args, local_vars, is_async=is_async
        )


def _reconcile_types(type_a, type_b):
    if type_a is Signature.empty:
        return type_b
    if type_b is Signature.empty:
        return type_a
    if type_a is type_b:
        return type_a
    raise Exception(f"Unable to reconcile types {type_a!r} and {type_b!r}")



from inspect import Parameter, signature

from incant import Hook, Incanter


def test_simple_override(incanter: Incanter):
    incanter.register_by_type(lambda: 5, type=int)

    def fn(dep1: int):
        return dep1 + 1

    assert incanter.invoke(fn) == 6

    additional_hooks = [Hook.for_name("dep1", lambda: 0)]

    assert incanter.prepare(fn, additional_hooks)() == 1
    assert incanter.prepare(fn, additional_hooks)() == 1
    assert incanter.prepare(fn, additional_hooks)() == 1

    additional_hooks = [Hook.for_type(int, lambda: 10)]

    assert incanter.prepare(fn, additional_hooks)() == 11
    assert incanter.prepare(fn, additional_hooks)() == 11


def test_override_to_parameter(incanter: Incanter):
    """A dependency can be overriden to a parameter."""
    incanter.register_by_type(lambda: 5, type=int)

    def fn(dep1: int):
        return dep1 + 1

    assert incanter.invoke(fn) == 6

    additional_hooks = [Hook.for_name("dep1", None)]

    assert signature(incanter.prepare(fn, additional_hooks)).parameters == {
        "dep1": Parameter("dep1", Parameter.POSITIONAL_OR_KEYWORD, annotation=int)
    }
    assert signature(incanter.prepare(fn)).parameters == {}



"""Test scenarios from the wired library.

Note: this isn't necessarily idiomatic incant usage, just tests
to prove feature parity.
"""
from collections import defaultdict
from typing import List

from attr import Factory, define

from incant import Incanter


def test_greeter(incanter: Incanter):
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/simple/index.html"""

    @define
    class Greeter:
        greeting: str

        def __call__(self) -> str:
            return f"{self.greeting} !!"

    greeter = Greeter(greeting="Hello")
    incanter.register_hook(lambda p: p.annotation is Greeter, lambda: greeter)

    def greet_a_customer(greeter: Greeter) -> str:
        return greeter()

    assert incanter.invoke(greet_a_customer) == "Hello !!"


def test_greeter_factory(incanter: Incanter):
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/factory/index.html"""

    @define
    class Greeter:
        greeting: str

        def __call__(self) -> str:
            return f"{self.greeting} !!"

    incanter.register_hook(
        lambda p: p.annotation is Greeter, lambda: Greeter(greeting="Hello")
    )

    def greet_a_customer(greeter: Greeter) -> str:
        return greeter()

    assert incanter.invoke(greet_a_customer) == "Hello !!"


def test_greeter_settings():
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/settings/index.html"""

    @define
    class Settings:
        punctuation: str

    @define
    class Greeter:
        greeting: str
        punctuation: str

        def __call__(self) -> str:
            return f"{self.greeting} {self.punctuation}"

    def greet_a_customer(greeter: Greeter) -> str:
        return greeter()

    def setup(settings: Settings):
        incanter = Incanter()
        punctuation = settings.punctuation

        def greeter_factory() -> Greeter:
            return Greeter(greeting="Hello", punctuation=punctuation)

        incanter.register_hook(lambda p: p.annotation is Greeter, greeter_factory)
        return incanter

    settings = Settings(punctuation="!!")
    incanter = setup(settings)
    assert incanter.invoke(greet_a_customer) == "Hello !!"


def test_greeter_settings_idiomatic(incanter: Incanter):
    """
    Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/settings/index.html

    The example is rewritten to be idiomatic incanter.
    """

    @define
    class Greeter:
        greeting: str
        punctuation: str

        def __call__(self) -> str:
            return f"{self.greeting} {self.punctuation}"

    def greet_a_customer(greeter: Greeter) -> str:
        return greeter()

    @incanter.register_by_type
    def greeter_factory(punctuation: str) -> Greeter:
        return Greeter(greeting="Hello", punctuation=punctuation)

    incanter.register_hook(lambda p: p.name == "punctuation", lambda: "!!")

    assert incanter.invoke(greet_a_customer) == "Hello !!"


def test_greeter_contexts(incanter: Incanter):
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/context/index.html"""

    @define
    class Customer:
        name: str

    @define
    class FrenchCustomer(Customer):
        pass

    @define
    class Greeter:
        punctuation: str
        greeting: str = "Hello"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    @define
    class FrenchGreeter(Greeter):
        greeting: str = "Bonjour"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    def greet_a_customer(customer: Customer, greeter: Greeter):
        return greeter(customer)

    incanter.register_by_name(lambda: "!!", name="punctuation")
    incanter.register_by_type(
        lambda customer, punctuation: FrenchGreeter(punctuation)
        if isinstance(customer, FrenchCustomer)
        else Greeter(punctuation),
        Greeter,
    )

    customer = Customer(name="Mary")
    french_customer = FrenchCustomer(name="Henri")

    assert incanter.invoke(greet_a_customer, customer) == "Hello Mary !!"
    assert incanter.invoke(greet_a_customer, french_customer) == "Bonjour Henri !!"


def test_greeter_decoupled(incanter: Incanter):
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/decoupled/index.html"""

    # The first part of the app:
    @define
    class Customer:
        name: str

    @define
    class Greeter:
        punctuation: str
        greeting: str = "Hello"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    incanter.register_by_name(lambda: "!!", name="punctuation")

    customer_types_to_greeters = defaultdict(lambda: Greeter)  # type: ignore

    def greeter_factory(customer, punctuation) -> Greeter:
        return customer_types_to_greeters[type(customer)](punctuation)

    incanter.register_by_type(greeter_factory)

    def greet_a_customer(customer: Customer, greeter: Greeter):
        return greeter(customer)

    customer = Customer(name="Mary")
    assert incanter.invoke(greet_a_customer, customer) == "Hello Mary !!"

    # The second part of the app:
    @define
    class FrenchCustomer(Customer):
        pass

    @define
    class FrenchGreeter(Greeter):
        greeting: str = "Bonjour"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    customer_types_to_greeters[FrenchCustomer] = FrenchGreeter

    french_customer = FrenchCustomer(name="Henri")
    assert incanter.invoke(greet_a_customer, french_customer) == "Bonjour Henri !!"


def test_greeter_datastore(incanter: Incanter):
    """Test the wired scenario from https://wired.readthedocs.io/en/latest/tutorial/datastore/index.html"""
    # The first part of the app:
    @define
    class Customer:
        name: str

    @define
    class Datastore:
        customers: List[Customer] = Factory(list)

    @define
    class Greeter:
        punctuation: str
        greeting: str = "Hello"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    datastore = Datastore()
    incanter.register_by_name(lambda: "!!", name="punctuation")
    incanter.register_by_type(lambda: datastore, Datastore)

    customer_types_to_greeters = defaultdict(lambda: Greeter)  # type: ignore
    incanter.register_by_type(
        lambda customer, punctuation: customer_types_to_greeters[type(customer)](
            punctuation
        ),
        Greeter,
    )

    customer = Customer(name="Mary")
    datastore.customers.append(customer)

    def customer_interaction(customer: Customer, greeter: Greeter):
        return greeter(customer)

    # The second part of the app:
    @define
    class FrenchCustomer(Customer):
        pass

    @define
    class FrenchGreeter(Greeter):
        greeting: str = "Bonjour"

        def __call__(self, customer: Customer) -> str:
            return f"{self.greeting} {customer.name} {self.punctuation}"

    customer_types_to_greeters[FrenchCustomer] = FrenchGreeter

    french_customer = FrenchCustomer(name="Henri")

    def add_to_datastore(datastore: Datastore):
        datastore.customers.append(french_customer)

    incanter.invoke(add_to_datastore)

    def sample_interactions(incanter: Incanter) -> List[str]:
        """Pretend to do a couple of customer interactions"""

        greetings = []

        def get_customers(datastore: Datastore) -> List[Customer]:
            return list(datastore.customers)

        customers = incanter.invoke(get_customers)

        for customer in customers:
            greeting = incanter.invoke(customer_interaction, customer)
            greetings.append(greeting)

        return greetings

    assert sample_interactions(incanter) == ["Hello Mary !!", "Bonjour Henri !!"]



import linecache

from contextlib import AbstractAsyncContextManager
from inspect import (
    Signature,
    isasyncgenfunction,
    iscoroutinefunction,
    signature,
)
from typing import Any, Callable, List, Union

from attr import define


def _is_async_context_manager(fn: Any) -> bool:
    return (
        fn.__class__ is type and issubclass(fn, AbstractAsyncContextManager)
    ) or bool(
        (wrapped := getattr(fn, "__wrapped__", None)) and isasyncgenfunction(wrapped)
    )


@define
class ParameterDep:
    arg_name: str
    type: Any
    default: Any = Signature.empty


@define
class LocalVarFactory:
    factory: Callable
    args: List[Union[Callable, ParameterDep]]


def compile_invoke(
    fn: Callable,
    fn_factory_args: List[str],
    outer_args: List[ParameterDep],
    local_vars: List[LocalVarFactory],
    is_async: bool = False,
) -> Callable:
    # Some arguments need to be taken from outside.
    # Some arguments need to be calculated from factories.
    sig = signature(fn)
    fn_name = f"invoke_{fn.__name__}" if fn.__name__ != "<lambda>" else "invoke_lambda"
    globs = {"_incant_inner_fn": fn}
    arg_lines = []
    for dep in outer_args:
        if dep.type is not Signature.empty:
            arg_type_snippet = f": _incant_arg_{dep.arg_name}"
            globs[f"_incant_arg_{dep.arg_name}"] = dep.type
        else:
            arg_type_snippet = ""
        if dep.default is not Signature.empty:
            arg_default = f"_incant_default_{dep.arg_name}"
            arg_type_snippet = f"{arg_type_snippet} = {arg_default}"
            globs[arg_default] = dep.default

        arg_lines.append(f"{dep.arg_name}{arg_type_snippet}")
    outer_arg_names = {o.arg_name for o in outer_args}

    lines = []

    ret_type = ""
    if sig.return_annotation is not Signature.empty:
        globs["_incant_return_type"] = sig.return_annotation
        ret_type = " -> _incant_return_type"
    if is_async:
        lines.append(f"async def {fn_name}({', '.join(arg_lines)}){ret_type}:")
    else:
        lines.append(f"def {fn_name}({', '.join(arg_lines)}){ret_type}:")
    local_vars_ix_by_factory = {
        local_var.factory: ix for ix, local_var in enumerate(local_vars)
    }
    ind = 0  # Indentation level
    for i, local_var in enumerate(local_vars):
        local_name = f"_incant_local_{i}"
        local_var_factory = f"_incant_local_factory_{i}"
        globs[local_var_factory] = local_var.factory
        local_arg_lines = []
        for local_arg in local_var.args:
            if isinstance(local_arg, ParameterDep):
                local_arg_lines.append(local_arg.arg_name)
            else:
                local_arg_lines.append(
                    f"_incant_local_{local_vars_ix_by_factory[local_arg]}"
                )
        if _is_async_context_manager(local_var.factory):
            lines.append(
                f"  {' ' * ind}async with {local_var_factory}({', '.join(local_arg_lines)}) as {local_name}:"
            )
            ind += 2
        else:
            aw = ""
            if iscoroutinefunction(local_var.factory):
                aw = "await "
            lines.append(
                f"  {' ' * ind}{local_name} = {aw}{local_var_factory}({', '.join(local_arg_lines)})"
            )

    incant_arg_lines = []
    local_var_ix = len(local_vars) - 1
    for name in sig.parameters:
        if name not in fn_factory_args and name in outer_arg_names:
            incant_arg_lines.append(name)
        else:
            incant_arg_lines.append(f"_incant_local_{local_var_ix}")
            local_var_ix -= 1

    aw = ""
    if iscoroutinefunction(fn):
        aw = "await "
    lines.append(
        f"  {' ' * ind}return {aw}_incant_inner_fn({', '.join(incant_arg_lines)})"
    )

    script = "\n".join(lines)

    fname = _generate_unique_filename(fn.__name__, "invoke", lines)
    eval(compile(script, fname, "exec"), globs)

    fn = globs[fn_name]
    return fn


def compile_incant_wrapper(
    fn: Callable, incant_plan: List[Union[int, str]], num_pos_args: int, num_kwargs: int
):
    fn_name = f"incant_{fn.__name__}" if fn.__name__ != "<lambda>" else "incant_lambda"
    globs = {"_incant_inner_fn": fn}
    arg_lines = []
    if incant_plan:
        if num_pos_args:
            arg_lines.append("*args")
        else:
            arg_lines.append("*")

    kwargs = [arg for arg in incant_plan if isinstance(arg, str)]
    arg_lines.extend(kwargs)
    if num_kwargs > len(kwargs):
        arg_lines.append("**kwargs")

    lines = []
    lines.append(f"def {fn_name}({', '.join(arg_lines)}):")
    lines.append("  return _incant_inner_fn(")
    for arg in incant_plan:
        if isinstance(arg, int):
            lines.append(f"    args[{arg}],")
        else:
            lines.append(f"    {arg},")
    lines.append("  )")

    script = "\n".join(lines)

    fname = _generate_unique_filename(fn.__name__, "incant", lines)
    eval(compile(script, fname, "exec"), globs)

    fn = globs[fn_name]
    return fn


def _generate_unique_filename(func_name: str, func_type: str, source: List[str]):
    """
    Create a "filename" suitable for a function being generated.
    """
    extra = ""
    count = 1

    while True:
        unique_filename = f"<incant generated {func_type} of {func_name}{extra}>"
        # To handle concurrency we essentially "reserve" our spot in
        # the linecache with a dummy line.  The caller can then
        # set this value correctly.
        cache_line = (len(source), None, source, unique_filename)
        if linecache.cache.setdefault(unique_filename, cache_line) == cache_line:  # type: ignore
            return unique_filename

        # Looks like this spot is taken. Try again.
        count += 1
        extra = f"-{count}"



import pytest

from incant import Incanter


def test_no_args(incanter: Incanter):
    def func():
        return 1

    assert incanter.incant(func) == 1


def test_no_args_extra_args(incanter: Incanter):
    """Unnecessary arguments are ignored."""

    def func():
        return 1

    assert incanter.incant(func, a=1, b=2)


def test_simple_args(incanter: Incanter):
    def func(a: int, b: int):
        return a + b

    assert incanter.incant(func, a=1, b=1) == 2


def test_pos_args_by_type(incanter: Incanter):
    def func(x: int) -> int:
        return x + 1

    assert incanter.incant(func, 5) == 6


@pytest.mark.asyncio
async def test_async_pos_args_by_type(incanter: Incanter):
    async def func(x: int) -> int:
        return x + 1

    assert await incanter.aincant(func, 5) == 6


def test_missing_args(incanter: Incanter):
    def func(x: int) -> int:
        return x + 1

    with pytest.raises(TypeError):
        incanter.incant(func)



from functools import wraps
from inspect import Parameter
from ipaddress import IPv4Address
from typing import NewType

from attrs import define, has
from cattrs import structure
from quart import Quart, request
from quattro import TaskGroup
from structlog.stdlib import BoundLogger, get_logger
from werkzeug.exceptions import BadRequest

from incant import Incanter


app = Quart(__name__)
incanter = Incanter()
logger = get_logger()


Header = NewType("Header", str)


def make_attrs_payload_factory(attrs_cls: type):
    async def attrs_payload_factory():
        json = await request.get_json(force=True)
        try:
            return structure(json, attrs_cls)
        except Exception as e:
            raise BadRequest() from e

    return attrs_payload_factory


incanter.register_hook_factory(
    lambda p: has(p.annotation), lambda p: make_attrs_payload_factory(p.annotation)
)
incanter.register_by_type(TaskGroup)


@incanter.register_by_type
def get_ip_address() -> IPv4Address:
    # In Quart (like in Flask), the request is accessed through a global proxy
    return IPv4Address(request.remote_addr)


@define
class User:
    """The user model."""

    username: str


async def current_user(session_token: str) -> User:
    # Complex black magic goes here, immune to timing attacks.
    return User("admin")


incanter.register_by_type(current_user)
incanter.register_by_name(
    lambda: request.cookies["session_token"], name="session_token"
)


def make_header_factory(name: str, default):
    if default is Parameter.empty:
        return lambda: request.headers[name.replace("_", "-")]
    else:
        return lambda: request.headers.get(name.replace("_", "-"), default)


incanter.register_hook_factory(
    lambda p: p.annotation is Header, lambda p: make_header_factory(p.name, p.default)
)


def quickapi(handler):
    log = logger.bind(handler=handler.__name__)

    prepared = incanter.prepare(handler, is_async=True)

    @wraps(handler)
    async def wrapper(**kwargs):
        log.info("Processing")
        return await incanter.aincant(prepared, log=log, **kwargs)

    return wrapper


@app.get("/")
@quickapi
async def index():
    return "OK"


@app.get("/ip")
@quickapi
async def ip_address_handler(source_ip: IPv4Address) -> str:
    return f"Your address is {source_ip}"


@app.get("/log")
@quickapi
async def logging_handler(log: BoundLogger) -> str:
    log.info("Hello from the log handler")
    return "Response after logging"


@app.get("/even-or-odd/<int:integer>")
@quickapi
async def even_or_odd_handler(integer: int) -> str:
    return "odd" if integer % 2 != 0 else "even"


@app.get("/user")
@quickapi
async def user_handler(user: User, log) -> str:
    log.info("Chilling here", user=repr(user))
    return "After the user handler"


@define
class SamplePayload:
    field: int


@app.post("/payload")
@quickapi
async def attrs_handler(payload: SamplePayload, log: BoundLogger) -> str:
    log.info("Received payload", payload=repr(payload))
    return "After payload"


@app.get("/taskgroup")
@quickapi
async def taskgroup_handler(tg: TaskGroup, log: BoundLogger) -> str:
    async def inner():
        log.info("Using structured concurrency, not leaking tasks")

    tg.create_task(inner())
    return "nice"


@app.get("/header")
@quickapi
async def a_header_handler(content_type: Header = Header("none"), log=logger) -> str:
    return f"The header was: {content_type}"


if __name__ == "__main__":
    app.run()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['incant']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

extras_require = \
{'dev': ['black',
         'cattrs',
         'coverage',
         'flake8',
         'httpx',
         'isort',
         'mypy',
         'pytest-asyncio',
         'pytest',
         'quart',
         'quattro',
         'rich',
         'structlog',
         'uvicorn']}

setup(name='incant',
      version='0.3',
      description='Magical function composition',
      author=None,
      author_email='Tin Tvrtkovic <tinchester@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



from inspect import Parameter, Signature, signature

from incant import Incanter


def test_params_with_defaults():
    """Default values should be propagated properly."""
    incant = Incanter()

    def fn(default_param=1):
        return default_param

    assert incant.invoke(fn) == 1

    assert signature(incant.prepare(fn)) == Signature(
        [Parameter("default_param", Parameter.POSITIONAL_OR_KEYWORD, default=1)]
    )



from asyncio import sleep
from collections import OrderedDict
from contextlib import asynccontextmanager
from inspect import Parameter, signature

import pytest

from quattro import TaskGroup

from incant import Incanter


@pytest.mark.asyncio
async def test_async_invoke(incanter: Incanter):
    async def fn():
        await sleep(0.001)
        return 2

    assert (await incanter.ainvoke(fn)) == 2


@pytest.mark.asyncio
async def test_async_dep(incanter: Incanter):
    @incanter.register_by_name
    async def dep1() -> int:
        return 1

    with pytest.raises(Exception):
        incanter.invoke(lambda dep1: dep1 + 1) == 2

    assert (await incanter.ainvoke(lambda dep1: dep1 + 1)) == 2
    assert signature(
        incanter.prepare(lambda dep1: dep1 + 1, is_async=True)
    ).parameters == OrderedDict([])


@pytest.mark.asyncio
async def test_async_mixed_dep(incanter: Incanter):
    @incanter.register_by_name
    async def dep1(dep2) -> int:
        return dep2 + 1

    @incanter.register_by_name(name="dep2")
    def _(input: int) -> int:
        return input + 1

    with pytest.raises(Exception):
        incanter.invoke(lambda dep1: dep1 + 1, 1)

    assert (await incanter.ainvoke(lambda dep1: dep1 + 1, 1)) == 4
    assert signature(
        incanter.prepare(lambda dep1: dep1 + 1, is_async=True)
    ).parameters == OrderedDict(
        [("input", Parameter("input", Parameter.POSITIONAL_OR_KEYWORD, annotation=int))]
    )


@pytest.mark.asyncio
async def test_async_ctx_manager_dep(incanter: Incanter):
    """Async context manager dependencies work."""
    entered, exited = False, False

    @incanter.register_by_name
    @asynccontextmanager
    async def dep1():
        nonlocal entered, exited
        entered = True
        yield 1
        exited = True

    async def fn(dep1: int) -> int:
        nonlocal entered
        return dep1 + 1

    assert (await incanter.ainvoke(fn)) == 2

    assert entered
    assert exited


@pytest.mark.asyncio
async def test_taskgroup_dep(incanter: Incanter):
    """Async context manager dependencies work."""
    incanter.register_by_type(TaskGroup)

    async def fn(tg: TaskGroup):
        return 2

    assert (await incanter.ainvoke(fn)) == 2


def test_async_invoke_return_type(incanter: Incanter):
    """Async context manager dependencies work."""

    async def fn() -> int:
        return 2

    assert signature(incanter.prepare(fn)).return_annotation is int



"""Tests for the quickapi module."""
from asyncio import create_task, sleep

import pytest

from httpx import AsyncClient
from uvicorn import Config, Server  # type: ignore

from .quickapi import app


class UvicornTestServer(Server):
    def install_signal_handlers(self) -> None:
        return


@pytest.fixture
async def quickapi_server(unused_tcp_port_factory):
    port = unused_tcp_port_factory()
    s = UvicornTestServer(Config(app=app, port=port))
    create_task(s.serve())
    yield f"http://localhost:{port}"
    s.should_exit = True
    await sleep(0.2)


@pytest.mark.asyncio
async def test_index(quickapi_server: str):
    async with AsyncClient() as client:
        resp = (await client.get(f"{quickapi_server}/")).text
        assert resp == "OK"


@pytest.mark.asyncio
async def test_async_context_manager(quickapi_server: str):
    async with AsyncClient() as client:
        resp = (await client.get(f"{quickapi_server}/taskgroup")).text
        assert resp == "nice"


@pytest.mark.asyncio
async def test_payload_handler(quickapi_server: str):
    async with AsyncClient() as client:
        resp = (await client.get(f"{quickapi_server}/header")).text
        assert resp == "The header was: none"

        resp = (
            await client.post(f"{quickapi_server}/payload", content=b'{"field": 1}')
        ).text
        assert resp == "After payload"


@pytest.mark.asyncio
async def test_header_handler(quickapi_server: str):
    async with AsyncClient() as client:
        resp = (await client.get(f"{quickapi_server}/header")).text
        assert resp == "The header was: none"

        resp = (
            await client.get(
                f"{quickapi_server}/header", headers={"content-type": "test"}
            )
        ).text
        assert resp == "The header was: test"
